@@ -25,8 +25,8 @@ class ThrowsTag extends AbstractTypeableTag implements TagInterface
     public function generate()
     {
         $output = '@throws'
-        . ((!empty($this->types)) ? ' ' . $this->getTypesAsString() : '')
-        . ((!empty($this->description)) ? ' ' . $this->description : '');
+        . (! empty($this->types) ? ' ' . $this->getTypesAsString() : '')
+        . (! empty($this->description) ? ' ' . $this->description : '');
 
         return $output;
     }
